#include "ipv4.h"

#define ANSI_COLOR_RED     "\x1b[31m"
#define ANSI_COLOR_GREEN   "\x1b[32m"
#define ANSI_COLOR_YELLOW  "\x1b[33m"
#define ANSI_COLOR_BLUE    "\x1b[34m"
#define ANSI_COLOR_MAGENTA "\x1b[35m"
#define ANSI_COLOR_CYAN    "\x1b[36m"
#define ANSI_COLOR_RESET   "\x1b[0m"

char * ipntoa
(const uint32_t ip, char * dst, const int length)
{
  unsigned char oct[4];

  if (length < 16)
    {
      fprintf(stderr, "ipntoa: buffer must be at least 16 bits wide.\n");
      exit(-1);
    }

  if (dst == NULL)
    {
      dst = malloc(sizeof(char) * 16);
    }

  memset(dst, 0, length);
  oct[0] = (unsigned char)(ip >> 24);
  oct[1] = (unsigned char)(ip >> 16);
  oct[2] = (unsigned char)(ip >> 8);
  oct[3] = (unsigned char)(ip & 0x000000ff);

  if (snprintf(dst, length, "%d.%d.%d.%d", oct[0], oct[1], oct[2], oct[3]) == length)
    {
      fprintf(stderr, "Warning: ipntoa exceeded buffer size.\n");
    }
  return(dst);
}

uint32_t networkNumber
(uint32_t ip, uint32_t sm)
{
    return (ip & sm);
}

uint32_t subnetDirectedBroadcast
(uint32_t ip, uint32_t sm)
{
    return (ip | ~sm);
}

uint32_t wildcardMask
(uint32_t sm)
{
    return ~sm;
}

char classful
(const uint32_t ip)
{
    char class = 0;
    if (ip >= 0xf0000000)      /* High bits = 1111 */
        class = 'E';
    else if (ip >= 0xe0000000) /* High bits = 1110 */
        class = 'D';
    else if (ip >= 0xc0000000) /* High bits = 110  */
        class = 'C';
    else if (ip >= 0x80000000) /* High bits = 10   */
        class = 'B';
    else                       /* High bit  = 0    */
        class = 'A';
    return(class);
}

uint32_t classfulMask
(const uint32_t ip)
{
    uint32_t mask = 0;
    if (ip >= 0xc0000000)       /* High bits = 110 */
        mask = 0xffffff00;
    else if (ip >=  0x80000000) /* High bits = 10  */
        mask = 0xffff0000;
    else                        /* High bit  = 0   */
        mask = 0xff000000;
    return(mask);
}

uint32_t summaryAddress
(uint32_t ip1, uint32_t sm1, uint32_t ip2, uint32_t sm2)
{
    return (networkNumber(ip1, sm1) & networkNumber(ip2, sm2));
}

uint32_t summaryPrefix
(uint32_t ip1, uint32_t sm1, uint32_t ip2, uint32_t sm2)
{
  uint32_t oldpfx, pfx, diff, pfxlen, network1, network2;
#ifdef DEBUG_SUMMARY_PREFIX
  char * diffip, * pfxold, * pfxnew, * ip1_s, * ip2_s;
  printf(ANSI_COLOR_BLUE);
  printf("                 SUMMARY PREFIX\n");
#endif

  oldpfx = (sm1 & sm2);
  network1 = networkNumber(ip1, sm1);
  network2 = networkNumber(ip2, sm2);

  /* Find the different bits in both networks.
     The extra look at the combined subnet mask
     catches a corner case where one network is
     a subnet of the other, but the different bits
     are further than either netmask. */
  diff = (network1 ^ network2) & oldpfx;

  /* Now there are four possibilities:
     1) The network numbers equal.
     2) Network2 is a subnet of Network1.
     3) Network1 is a subnet of Network2.
     4) Each network is distinct and can be summarized. */
  if (network1 == network2 && sm1 == sm2)
    {
      pfx = oldpfx;
#ifdef DEBUG_SUMMARY_PREFIX
      ip1_s = ipntoa(ip1, NULL, 16);
      ip2_s = ipntoa(ip2, NULL, 16);
      fprintf(stderr,
	      "No need to summarize route.\n%s and %s are in the same network.\n",
	      ip1_s, ip2_s);
      free(ip1_s);
      free(ip2_s);
#endif
    }
  else if (diff == 0 && sm1 < sm2)
    {
      pfx = sm1;
#ifdef DEBUG_SUMMARY_PREFIX
      ip1_s = ipntoa(network1, NULL, 16);
      ip2_s = ipntoa(network2, NULL, 16);
      fprintf(stderr, "No need to summarize route.\n%s/%u contains %s/%u.\n",
	      ip1_s, prefixLength(sm1), ip2_s, prefixLength(sm2));
      free(ip1_s);
      free(ip2_s);
#endif
    }
  else if (diff == 0 && sm1 > sm2)
    {
      pfx = sm2;
#ifdef DEBUG_SUMMARY_PREFIX
      ip1_s = ipntoa(network1, NULL, 16);
      ip2_s = ipntoa(network2, NULL, 16);
      fprintf(stderr, "No need to summarize route.\n%s/%u contains %s/%u.\n",
	      ip2_s, prefixLength(sm2), ip1_s, prefixLength(sm1));
      free(ip1_s);
      free(ip2_s);
#endif
    }
  else
    {
      pfxlen = (32 - __builtin_clz(diff) - __builtin_ctz(oldpfx));
      pfx = oldpfx << pfxlen;

#ifdef DEBUG_SUMMARY_PREFIX
      diffip = ipntoa(diff, NULL, 16);
      pfxold = ipntoa(oldpfx, NULL, 16);
      pfxnew = ipntoa(pfx, NULL, 16);

      printf("        Value             CLZ  CTZ\n");
      printf("pfxold: %-16s  %-3d  %d\n", pfxold, __builtin_clz(oldpfx), __builtin_ctz(oldpfx));
      printf("diff:   %-16s  %-3d  %d\n", diffip, __builtin_clz(diff), __builtin_ctz(diff));
      printf("pfxnew: %-16s  %-3d  %d\n", pfxnew, __builtin_clz(pfx), __builtin_ctz(pfx));
      free (diffip);
      free (pfxold);
      free (pfxnew);
#endif
    }
#ifdef DEBUG_SUMMARY_PREFIX
  printf(ANSI_COLOR_RESET);
#endif
  return(pfx);
}

uint32_t prefixLength
(const uint32_t sm)
{
  return ((33 - ffs(sm)) % 33); /* mod 33 to handle special 0 case */
}

uint32_t countHosts
(const uint32_t sm)
{
  uint32_t prefix, hosts;
  prefix = prefixLength(sm);
  if (sm > 0x00000000 && sm < 0xfffffffe) /* 1.0.0.0 - 255.255.255.252 */
    {
      hosts = (1 << (32 - prefix)) - 2;
    }
  else if (sm == 0xfffffffe)              /* 255.255.255.254 */
    {

    }
  else                                    /* 0.0.0.0 and 255.255.255.255 */
    {
      hosts = 0;
    }
  return(hosts);
}

uint32_t countSubnets
(const uint32_t ip, const uint32_t sm)
{
    uint32_t subnets = 0;
    uint32_t classful = classfulMask(ip);

    if (sm > classful)
    {
        subnets = (1 << (prefixLength(sm) - prefixLength(classful)));
    }

    return(subnets);
}

int isValidNetmask
(const uint32_t sm)
{
  uint32_t valid = 0;
  if (sm == 0)
    {
      fprintf(stderr,
	      "0.0.0.0 may not technically be a valid subnet mask.\n");
      valid = 1;
    }
#ifdef GCC_FFS
  else if (__builtin_ffs(sm) + __builtin_popcount(sm) - 1 == 32)
#else
  else if (ffs(sm) + __builtin_popcount(sm) - 1 == 32)
#endif
    {
      valid = 1;
    }
  else
    {
      char * str = calloc(sizeof(char), 16);
      fprintf(stderr,
	      ANSI_COLOR_RED
	      "%s is not a valid subnet mask. FFS=%-2d POP=%-2d\n"
	      ANSI_COLOR_RESET,
	      ipntoa(sm, str, 16),
#ifdef GCC_FFS
	      __builtin_ffs(sm) - 1,
#else
	      ffs(sm) - 1,
#endif
	      __builtin_popcount(sm));
      free(str);
      exit(-1);
    }
  return (valid);
}

uint32_t parse
(const char * src)
{
  uint32_t ip, i, x;
  char * token;
  char s[16];
  const char delim[2] = ".";

  strncpy(s, src, 16);

  if (strlen(s) > 16)
    {
      fprintf(stderr, ANSI_COLOR_RED "%s is not a valid IP.\n"
	      ANSI_COLOR_RESET, s);
      exit(-1);
    }
  else /* check for characters other than 0-9 and '.' */
    {
      int a;
      for (a = 0 ; a < (char) strlen(src) ; a++)
	{
	  if (s[a] != '.' && (s[a] < '0' || s[a] > '9'))
	    {
	      fprintf(stderr, ANSI_COLOR_RED
		      "%s is not a valid IP (invalid character: %c).\n"
		      ANSI_COLOR_RESET, src, s[a]);
	      exit(-1);
	    }
	}
    }

  ip = 0;
  i = 32;

  token = strtok(s, delim);
  while (i > 0 && token != NULL)
    {
      i -= 8;
      x = (uint32_t) atoi(token);
      
      if (x > 255)
	{
	  fprintf(stderr, ANSI_COLOR_RED
		  "%s is not a valid IP (invalid value: %s)\n"
		  ANSI_COLOR_RESET, src, token);
	  exit(-1);
	} 

      ip += (x << i);
      token = strtok(NULL, delim);
    }
  return(ip);
}

void generateSubnets
()
{
  uint32_t mask, i;
  char ip[16];
  mask = 0x00000000;
  for (i=0; i<33; i++) /* though ugly, 33 makes sure we catch case 0xffffffff */
    {
      ipntoa(mask, ip, 16);
      printf("%-2d\t%s\n", prefixLength(mask), ip);
      mask = (mask >> 1) + 0x80000000;
    }
}
